//temporalRow.hxx
//
//Chris Bentzel, 01.16.98

/***************************************************************************
 * 
 * Copyright 1998  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

#ifndef _TEMPORAL_ROW_HDR_
#define _TEMPORAL_ROW_HDR_

#include "temporalRowEntry.hxx"
#include "../include/genericException.hxx"
#include <vector.h>

//index exception classes... probably want to move these
class badIndexException : public genericException {
public:

  badIndexException() : genericException() {}
  badIndexException(string exceptionInfo) : genericException(exceptionInfo) {}
};

class negativeIndexException : public badIndexException {
public:
  
  negativeIndexException() : badIndexException() {}
  negativeIndexException(string exceptionInfo) : badIndexException(exceptionInfo) {}
};

class overflowIndexException : public badIndexException {
public:
  
  overflowIndexException() : badIndexException() {}
  overflowIndexException(string exceptionInfo) : badIndexException(exceptionInfo) {}
};


//mostly a container class, also returns supplier info
class temporalRow {

public:

  temporalRow(int supplierIndex =-1, int numberElements = 0);
  ~temporalRow();

  void getEntry(int index, temporalRowEntry& entryData) const throw(negativeIndexException, overflowIndexException);
  //getEntry():
  //          If index less than zero or >= number of temporalRowEntry elements,
  //                        badIndexException is thrown
  //
  //          If index is valid, copies the data of that entry to entryData
  
  void setEntry(int index, temporalRowEntry& entryData) throw(negativeIndexException, overflowIndexException);
  //setEntry():
  //          If index less than zero or >= number of temporalRowEntry elements,
  //                        badIndexException is thrown
  //
  //          If valid, copies data from entryData to appropriate entry

  void setLatestEntry(temporalRowEntry& entryData) throw(negativeIndexException);
  //setLatestEntry():
  //          Copies entryData to latest entry, as well as shifting the rest of the
  //          data one slot back (if necessary)
  //
  //          Throws badIndexException if 0 entries have been specified
  
  void getLatestEntry(temporalRowEntry& entryData) const throw(negativeIndexException);
  //getLatestEntry():
  //          Copies latest entry to entryData
  //
  //          Throws badIndexException if no entries have been specified
  
  int getSupplierIndex() const throw(negativeIndexException);
  //getSupplierIndex:
  //    if local index unset (i.e. has a value of -1), throws badIndexException
  //    otherwise, returns value
  
  void setSupplierIndex(int index) throw(negativeIndexException, overflowIndexException);
  //setSupplierIndex assumes that the index value is valid if positive
  // throws badIndexException if index < 0

  void setNumberElements(int numberElements) throw(negativeIndexException);
  //sets the number of elements for the vector
  //If the numberOfElements is less than current (but greater than zero) nothing happens
  //If numberElements < 0, throws badIndexException
  //If the numberOfElements greater than current, allocates new space

  int getNumberElements(void) const;
  //returns the number of elements in the row

private:
  
  //don't want assignments or copy constructor
  temporalRow(const temporalRow& tR);
  temporalRow& operator=(const temporalRow& tR);

  //helper functions
  void shift();
  
  //data
  int numElems;
  vector<temporalRowEntry> rep;
  int sup;
  int latest;
  
};



















//inlines

inline
temporalRow::temporalRow(int supplierIndex, int numberElements)
{
  sup = supplierIndex;
  numElems = numberElements;
  rep.reserve(numElems);
  latest = -1;
}

inline
temporalRow::~temporalRow()
{
}

inline int
temporalRow::getNumberElements(void) const
{
  return numElems;
}

inline void
temporalRow::shift()
{
  int lastElem = numElems - 1;
  for (int i = 0; i < lastElem; i++)
    {
      rep[i] = rep[i+1];
    }
}

#endif //_TEMPORAL_ROW_HDR_



















