//comTracker.c++
//Chris Bentzel
//Sep. 97

/***************************************************************************
 * 
 * Copyright 1997  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

/////////////////////////////////////////////////////////////////
////
////   Pre-processor directives and includes
////
/////////////////////////////////////////////////////////////////
#include "CTracker.h++"
#include "pvm3.h"
#include "GRAYimage.h++"
#include "messageTags.h++"
#include "CVisStateHandler.h++"
#include "mytime.h++"

#include <iostream.h>
#include <unistd.h>

//#define COM_TRACKER_DEBUG
//#define COM_TRACKER_NO_BACKSUB


/////////////////////////////////////////////////////////////////
////
////  Externally defined globals
////
/////////////////////////////////////////////////////////////////
extern CVisState* theCurrentState;
extern CVisState* theBackupState;
extern CTracker* theTracker;
extern int trackMainQuit;
extern int numberOfImages;

/////////////////////////////////////////////////////////////////
////
////  Variables to time data transfer rates (for debugging)
////
/////////////////////////////////////////////////////////////////
#ifdef COM_TRACKER_DEBUG

#ifndef COM_TRACKER_NO_BACKSUB
static long sendBitmapTime;
static long receiveBitmapTime;
#endif

static long sendStateTime;
static long receiveStateTime;
#endif



/////////////////////////////////////////////////////////////////
////
////  Globals useful for communication routines
////
/////////////////////////////////////////////////////////////////
static GRAYimage Im(320, 240);

//Bootstraps: Do initial call to visState and backsub if the
//modules exist and the bootstrap values are 0

#ifndef COM_TRACKER_NO_BACKSUB
static int backsubRequestBootstrap = 1;
#endif

static int visStateRequestBootstrap = 1;

//IDs: holder of IDs
int visStateID = PvmNoGroup;

#ifndef COM_TRACKER_NO_BACKSUB
int backsubID = PvmNoGroup;
#endif

//IDCounters, MaxIDCounters:
//Use old value of module ID until counter = maxIDNumber
//value of zero is bootstrap
static const int visStateMaxIDCounter = 3000;
static int visStateIDCounter = visStateMaxIDCounter;

#ifndef COM_TRACKER_NO_BACKSUB
static const int backsubMaxIDCounter = 3000;
static int backsubIDCounter = backsubMaxIDCounter;
#endif



#ifndef COM_TRACKER_NO_BACKSUB
/////////////////////////////////////////////////////////////////
////
////  handleTrackerImage()
////
/////////////////////////////////////////////////////////////////
void
handleTrackerImage(int theID)
{

#ifdef COM_TRACKER_DEBUG
  cerr << "Processing image reply\n";
#endif
  
#ifdef COM_TRACKER_DEBUG
  //report the time between sending an image request and receiving an image
  receiveBitmapTime = GetTimeNowMiliSeconds();
  cerr << "Image time difference : " << (receiveBitmapTime - sendBitmapTime) << endl;
#endif

  //Request another image before processing the current one.
  //Speeds up processing time, and guarantees max of two images
  pvm_initsend(PvmDataDefault);
  pvm_send(theID, BITMAP_REQUEST_MSGTAG);

#ifdef COM_TRACKER_DEBUG
  sendBitmapTime = GetTimeNowMiliSeconds();
#endif

  //Grab the new image and unpack it
  pvm_recv(theID, BITMAP_SENT_MSGTAG);
  Im.PVMunpack();
  
  //process new image and finish
  theTracker->addNewImage(Im);

  numberOfImages++;

#ifdef COM_TRACKER_DEBUG
  cerr << "Processed image reply\n";
#endif
  
}
#endif


/////////////////////////////////////////////////////////////////
////
////  handleTrackerRead()
////
/////////////////////////////////////////////////////////////////
void
handleTrackerRead(int theID)
{

#ifdef COM_TRACKER_DEBUG
  cerr << "Processing read reply\n";
#endif
  
#ifdef COM_TRACKER_DEBUG
  receiveStateTime = GetTimeNowMiliSeconds();
  cerr << "State time difference: "
       << (receiveStateTime - sendStateTime) << endl;
#endif

  //send out a request to read a new state
  pvm_initsend(PvmDataDefault);
  pvm_send(theID, VISSTATE_READ_REQUEST_MSGTAG);

#ifdef COM_TRACKER_DEBUG
  sendStateTime = GetTimeNowMiliSeconds();
#endif
  
  //receive the message and take it off the queue
  pvm_recv(theID, VISSTATE_READ_REPLY_MSGTAG);

  //Find out whether the read operation was a success or not
  int readResult;
  pvm_upkint(&readResult, 1, 1);

  //If successful, update the new state, and copy it over to
  //the backup state
  if (readResult == VIS_STATE_READ)
    {
      theCurrentState->pvmUnpack();
      theCurrentState->copy(theBackupState);
    }

  //If unsuccessful, copy the backup state to the new state
  //and try again later
  else
    {
      theBackupState->copy(theCurrentState);
    }

#ifdef COM_TRACKER_DEBUG
  cerr << "Processed read reply\n";
#endif
}



/////////////////////////////////////////////////////////////////
////
////  handleTrackerReadAndLock()
////
/////////////////////////////////////////////////////////////////
void
handleTrackerReadAndLock(int theID)
{

#ifdef COM_TRACKER_DEBUG
  cerr << "Processing Read And Lock Reply\n";
#endif
  
   //receive the message
  pvm_recv(theID, VISSTATE_READ_AND_LOCK_REPLY_MSGTAG);
  
  //see what the result of the read and lock request is
  int readAndLockResult;
  pvm_upkint(&readAndLockResult, 1, 1);
  
  //If it was successful, indicate that we have locked the state
  if (readAndLockResult == VIS_STATE_READ)
    {
      theCurrentState->pvmUnpack();
      theCurrentState->copy(theBackupState);
    }
  
  //If it wasn't successful, do nothin
  else
    {
    }

#ifdef COM_TRACKER_DEBUG
  cerr << "Processed Read and Lock Reply\n";
#endif
  
}



/////////////////////////////////////////////////////////////////
////
////  handleTrackerModify()
////
/////////////////////////////////////////////////////////////////
void
handleTrackerModify(int theID)
{

#ifdef COM_TRACKER_DEBUG
  cerr << "Processing modify reply\n";
#endif

  //receive the message
  pvm_recv(theID, VISSTATE_MODIFY_REPLY_MSGTAG);

#ifdef COM_TRACKER_DEBUG
  cerr << "received image\n";
#endif
  
  //See what the result of the operation was
  int modifyResult;
  pvm_upkint(&modifyResult, 1, 1);
  
  //If successfuly modified, see who modified it
  if (modifyResult == VIS_STATE_MODIFIED)
    {
      //Get the modifier
      int modifySource;
      pvm_upkint(&modifySource, 1, 1);

      //If the modifier is not the tracker, send out a read request
      if (modifySource != VIS_TRACK_SOURCE)
        {

#ifdef COM_TRACKER_DEBUG
          cerr << "this one should NOT be called\n";
#endif
          pvm_initsend(PvmDataDefault);
          pvm_send(theID, VISSTATE_READ_REQUEST_MSGTAG);
        }
      
      //otherwise, copy over the new state and report that visState is unlocked
      else
        {

#ifdef COM_TRACKER_DEBUG
          cerr << "this one should be called\n";
#endif
          theCurrentState->copy(theBackupState);
          //pvm_initsend(PvmDataDefault);
          //pvm_send(theID, VISSTATE_READ_REQUEST_MSGTAG);
        }
    }
  
  //if unsuccessful, copy the backup state to the new state
  else
    theBackupState->copy(theCurrentState);

#ifdef COM_TRACKER_DEBUG
  cerr << "Processed modify reply\n";
#endif

}

/////////////////////////////////////////////////////////////////
////
////  handleBitmapRequest()
////
/////////////////////////////////////////////////////////////////
void
handleBitmapRequest(int theID)
{

#ifdef COM_TRACKER_DEBUG
  cerr << "Processing image request\n";
#endif

  //receive the message
  pvm_recv(theID, BITMAP_REQUEST_MSGTAG);

#ifdef COM_TRACKER_DEBUG
  cerr << "Received message\n";
#endif
  
  pvm_initsend(PvmDataDefault);

  
  theTracker->sendImageSection();
  pvm_send(theID, BITMAP_SENT_MSGTAG);

#ifdef COM_TRACKER_DEBUG
  cerr << "Processed modify reply\n";
#endif

}



/////////////////////////////////////////////////////////////////
////
////  checkTrackerMessageQueue()
////
/////////////////////////////////////////////////////////////////
void
checkTrackerMessageQueue()
{

  //cerr << "checkTrackerMessageQueue\n\n";
  
#ifndef COM_TRACKER_NO_BACKSUB
  //Get the backsub ID again if it has been a while since the last
  //successful try
  if (backsubIDCounter == backsubMaxIDCounter)
    {
      //cerr << "trying to get backsubID\n";
      backsubID = pvm_gettid("backsub", 0);
      //cerr << "backsubID is : " << backsubID << "\n";
      //if successful in getting ID, restart counter
      if (backsubID > 0)
        backsubIDCounter = 0;
      //if unsuccessful, hold off for a little while and try again
      else
        backsubIDCounter = backsubMaxIDCounter - 20;
    }
#endif
  
  //Get the visState ID again if it has been a while since the
  //last successful try
  if (visStateIDCounter == visStateMaxIDCounter)
    {
      visStateID = pvm_gettid("visState", 0);
      //if successful in getting ID, restart counter
      if (visStateID > 0)
        visStateIDCounter = 0;
      //otherwise, hold off for a little while and try again
      else
        visStateIDCounter = visStateMaxIDCounter - 20;
    }

  //increment the counters
  
#ifndef COM_TRACKER_NO_BACKSUB
  backsubIDCounter++;
#endif

  visStateIDCounter++;
  
  //If neither of the ID's can be obtained...
  if (
      
#ifndef COM_TRACKER_NO_BACKSUB
      backsubID <=  0 &&
#endif
      
      visStateID <= 0)
    {
      //... because neither of the processes are registered with PVM,
      //return and try again
      if (

#ifndef COM_TRACKER_NO_BACKSUB
          backsubID == PvmNoGroup &&
#endif
          
          visStateID == PvmNoGroup)
        return;
      //... because there is a big PVM problem, quit the program
      else
        {
          trackMainQuit = 1;
          return;
        }
    }

     
  //Get the buffer ID's for various types of messages

#ifndef COM_TRACKER_NO_BACKSUB
  int bitmapSentBufID = pvm_probe(backsubID, BITMAP_SENT_MSGTAG);
#endif
  
  int readBufID = pvm_probe(visStateID, VISSTATE_READ_REPLY_MSGTAG);
  int readAndLockBufID = pvm_probe(visStateID, VISSTATE_READ_AND_LOCK_REPLY_MSGTAG);
  int modifyBufID = pvm_probe(visStateID, VISSTATE_MODIFY_REPLY_MSGTAG);
  int bitmapRequestBufID = pvm_probe(-1, BITMAP_REQUEST_MSGTAG);
  
  //No new messages have arrived
  if (

#ifndef COM_TRACKER_NO_BACKSUB
      bitmapSentBufID == 0 &&
#endif
      
      readBufID == 0 &&
      readAndLockBufID == 0 &&
      modifyBufID == 0 &&
      bitmapRequestBufID == 0)
    {
      //Sleep for a little while
      Sleep(25);
      
#ifndef COM_TRACKER_NO_BACKSUB
      //If the backsub module is up and needs to be bootstrapped
      if (backsubID > 0 && backsubRequestBootstrap)
        {

#ifdef COM_TRACKER_DEBUG
          cerr << "Bootstrapping backsub\n";
#endif
          //send a request for an image to the backsub module
          pvm_initsend(PvmDataDefault);
          pvm_send(backsubID, BITMAP_REQUEST_MSGTAG);
          backsubRequestBootstrap = 0;
        
#ifdef COM_TRACKER_DEBUG
          sendBitmapTime = GetTimeNowMiliSeconds();
#endif

        }
#endif
      
      //If the visState module is up and needs to be bootstrapped
      if (visStateID > 0 && visStateRequestBootstrap)
        {

#ifdef COM_TRACKER_DEBUG
          cerr << "Bootstrapping visState\n";
#endif
          //send out a request for the visState
          pvm_initsend(PvmDataDefault);
          pvm_send(visStateID, VISSTATE_READ_REQUEST_MSGTAG);
          visStateRequestBootstrap = 0;
          
#ifdef COM_TRACKER_DEBUG 
          sendStateTime = GetTimeNowMiliSeconds();
#endif
          
        }
      return;
    }
  
  //If a catastrophic error has occurred, quit the program
  if (

#ifndef COM_TRACKER_NO_BACKSUB
      bitmapSentBufID < 0 ||
#endif
      
      readBufID < 0 ||
      readAndLockBufID < 0 ||
      modifyBufID < 0 ||
      bitmapRequestBufID < 0)
    {
      trackMainQuit = 1;
      return;
    }

  //Deal with whichever message is in the queue

#ifndef COM_TRACKER_NO_BACKSUB
  if (bitmapSentBufID > 0)
    {

#ifdef COM_TRACKER_DEBUG
      int bytes, sender, msgtag;
      pvm_bufinfo(bitmapSentBufID, &bytes, &msgtag, &sender);
      cerr << "Message size: " << bytes << " bytes\n";
#endif

      handleTrackerImage(backsubID);
    }
#endif
  
  if (readBufID > 0)
    handleTrackerRead(visStateID);
  
  if (readAndLockBufID > 0)
    handleTrackerReadAndLock(visStateID);

  if (modifyBufID > 0)
    handleTrackerModify(visStateID);

  if (bitmapRequestBufID > 0)
    {
      int bytes, sender, msgtag;
      pvm_bufinfo(bitmapRequestBufID, &bytes, &msgtag, &sender);
      handleBitmapRequest(sender);

#ifdef COM_TRACKER_DEBUG
      cerr << "Sender is " << sender << endl;
#endif
      
    }
  
  return;
}





















